#include "scene.h"
#include "v3.h"
#include "m33.h"
#include "ppc.h"
#include "tm.h"
#include <stdlib.h>

Scene *scene;

using namespace std;

#include <fstream>

#include <iostream>

Scene::Scene() {

	ofs = 0;

	cgi = 0;
	soi = 0;

	L = V3(0.0f, 0.0f, 0.0f);

	tms = 0;
	tmsN = 0;

	gui = new GUI();
	gui->show();

	int u0 = 20;
	int v0 = 20;
//	int w = 640;
//	int h = 480;

	int w = 1000;
	int h = 600;

	fb = new FrameBuffer(u0, v0, w, h);
	fb->label("SW Framebuffer");
	fb->show();

	hwfb = new FrameBuffer(u0 + fb->w + 30, v0, w, h);
	hwfb->label("HW Framebuffer");
	hwfb->ishw = true;
//	hwfb->show();

	gpufb = new FrameBuffer(u0, v0, w, h);
	gpufb->label("GPU Framebuffer");
	gpufb->isgpu = true;
//	gpufb->show();

	fb3 = new FrameBuffer(u0 + fb->w + 30, v0, w, h);
	fb3->label("Third person");
//	fb3->show();

	float hfov = 105.0f;
//	float hfov = 55.0f;
	ppc = new PPC(hfov, fb->w, fb->h);
	ppc3 = new PPC(30.0f, fb3->w, fb3->h);

	gui->uiw->position(u0 + 20, v0 + fb->h+40);

	tmsN = 1;
	tms = new TM[tmsN];
	tms[0].LoadBin("geometry/teapot1K.bin");
	V3 tmC = ppc->C + ppc->GetVD()*100.0f;
	float tmSize = 100.0f;
	tms[0].PositionAndSize(tmC, tmSize);


	ppc3->C = ppc3->C + V3(330.0f, 150.0f, 300.0f);
	ppc3->PositionAndOrient(ppc3->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));

	ka = 0.2f;

//	Render();

	leftfb = new FrameBuffer(30, 40, 10, 10);
	leftfb->LoadTiff("images/left.tif");
	leftfb->label("Left");
	leftfb->show();

	rightfb = new FrameBuffer(30 + 30 + leftfb->w, 40, 10, 10);
	rightfb->LoadTiff("images/right.tif");
	rightfb->label("Right");
	rightfb->show();

	leftppc = new PPC(55.0f, 640, 480);
	rightppc = new PPC(55.0f, 640, 480);
	rightppc->Pan(-20.0f);

}

void Scene::RenderHW() {

	// clear the frame buffers
	glEnable(GL_DEPTH_TEST);

	glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// set camera instrinsics
	ppc->SetIntrinsicsHW();
	// set camera extrinsics
	ppc->SetExtrinsicsHW();

	// render geometry
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderHW();
	}

}



void Scene::RenderGPU() {

	// if the first time, call per session initialization
	if (cgi == NULL) {
		cgi = new CGInterface();
		cgi->PerSessionInit();
		soi = new ShaderOneInterface();
		soi->PerSessionInit(cgi);
	}

	// clear the framebuffer
	glEnable(GL_DEPTH_TEST);
	glClearColor(0.0, 0.0f, 0.5f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// set intrinsics
	ppc->SetIntrinsicsHW();
	// set extrinsics
	ppc->SetExtrinsicsHW();

	// per frame initialization
	cgi->EnableProfiles();
	soi->PerFrameInit();
	soi->BindPrograms();

	// render geometry
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderHW();
	}

	soi->PerFrameDisable();
	cgi->DisableProfiles();



}


void Scene::Render() {

	if (fb) {
		Render(ppc, fb);
	}

	float currf = 40.0f;
	if (fb3) {
		fb3->SetBGR(0xFFFFFFFF);
		fb3->ClearZB(0.0f);
		Render(ppc3, fb3);
		fb3->Draw3DPoint(L, V3(1.0f, 1.0f, 0.3f), ppc3, 7);
//		ppc->Visualize(ppc3, fb3, currf);
//		fb->Visualize(ppc, currf, ppc3, fb3);
//		fb->Visualize3D(ppc, currf, ppc3, fb3);
	}

}

void Scene::Render(PPC *currppc, FrameBuffer *currfb) {

	currfb->SetBGR(0xFFFFFFFF);
	currfb->ClearZB(0.0f);
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderFilled(currppc, currfb);
//		tms[tmi].RenderWireframe(currppc, currfb);
//		tms[tmi].RenderPoints(currppc, currfb);
	}
	currfb->redraw();

}

void Scene::AnalyzeDepthError(TM* tptm, V3 v, V3 vc, PPC* mppc,
	PPC* sppc, FrameBuffer* sfb) {

	V3 pn = ((tptm->verts[1] - tptm->verts[0]) ^ (tptm->verts[2] - tptm->verts[0])).UnitVector();
	V3 v0 = tptm->verts[0];
	V3 pi = v.IntersectRayWithPlane((v - mppc->C).UnitVector(), pn, v0);
	sfb->Draw3DPoint(v, V3(0.0f, 0.0f, 0.0f), sppc, 5);
	sfb->Draw3DPoint(pi, motionColor, sppc, 5);
	sfb->Draw3DSegment(v, V3(0.0f, 0.0f, 0.0f), pi, motionColor, sppc);
	V3 pv(FLT_MAX, FLT_MAX, FLT_MAX), ppi(FLT_MAX, FLT_MAX, FLT_MAX);
	sppc->Project(v, pv);
	sppc->Project(pi, ppi);
	pv[2] = ppi[2] = 0.0f;
	if (ofs) {
//		*ofs << v << " " << pi << " " << (v - pi).Length() << " " << (pv - ppi).Length() << endl;
		*ofs <<  (pv - ppi).Length() << " ";
	}
}

void Scene::AnalyzeDepthError(TM* tptm, TM* lowtm, TM* hightm, PPC* mppc, 
	PPC* sppc, FrameBuffer* sfb) {

//	*ofs << "Low depth" << endl;
	for (int vi = 0; vi < lowtm->vertsN; vi++) {
		AnalyzeDepthError(tptm, lowtm->verts[vi], lowtm->colors[vi], mppc, sppc, sfb);
	}
	AnalyzeDepthError(tptm, lowtm->GetCenter(), lowtm->colors[0], mppc, sppc, sfb);

//	*ofs << "High depth" << endl;
	for (int vi = 0; vi < hightm->vertsN; vi++) {
		AnalyzeDepthError(tptm, hightm->verts[vi], hightm->colors[vi], mppc, sppc, sfb);
	}
	AnalyzeDepthError(tptm, hightm->GetCenter(), hightm->colors[0], mppc, sppc, sfb);

	*ofs << endl;

}


void Scene::TestDepthSensitivity() {

	V3 bO(0.0f, 0.0f, -6.0f);
	V3 ws(1.0f, 1.0f, 1.0f);
	float zrange = 0.2f;

	TM bmtm;
	bmtm.SetRectangle(bO, ws[0], ws[2]);
	bmtm.RotateAboutArbitraryAxis(bO, V3(1.0f, 0.0f, 0.0f), 90.0f);
	bmtm.Translate(V3(0.0f, -1.5f, 0.0f));

	TM tptm;
	tptm.SetRectangle(bO, ws[0], ws[2]);
	tptm.RotateAboutArbitraryAxis(bO, V3(1.0f, 0.0f, 0.0f), 90.0f);
	tptm.Translate(V3(0.0f, -1.5f, 0.0f));
	tptm.Translate(V3(0.0f, 1.0f, 0.0f));

	TM hightm;
	hightm.SetRectangle(bO, ws[0], ws[2]);
	hightm.RotateAboutArbitraryAxis(bO, V3(1.0f, 0.0f, 0.0f), 90.0f);
	hightm.Translate(V3(0.0f, -1.5f, 0.0f));
	hightm.Translate(V3(0.0f, 1.0f+zrange, 0.0f));
	hightm.SetColor(0xFFEEEEEE);

	TM lowtm;
	lowtm.SetRectangle(bO, ws[0], ws[2]);
	lowtm.RotateAboutArbitraryAxis(bO, V3(1.0f, 0.0f, 0.0f), 90.0f);
	lowtm.Translate(V3(0.0f, -1.5f, 0.0f));
	lowtm.Translate(V3(0.0f, 1.0f-zrange, 0.0f));
	lowtm.SetColor(0xFFEEEEEE);


	FrameBuffer *mcamfb = new FrameBuffer(60, 60, 640, 480);
	mcamfb->label("Mentee camera");
	mcamfb->show();
	PPC *mppc = new PPC(55.0f, 640, 480);
	mppc->C = tptm.GetCenter() + V3(0.0f, 0.8f, ws[2]/2.0f+0.2f);
	mppc->PositionAndOrient(mppc->C, tptm.GetCenter(), V3(0.0f, 1.0f, 0.0f));

	mcamfb->SetBGR(0xFFFFFFFF);
	mcamfb->ClearZB(0.0f);
	tptm.RenderWireframe(mppc, mcamfb);
	lowtm.RenderWireframe(mppc, mcamfb);
	hightm.RenderWireframe(mppc, mcamfb);
	mcamfb->redraw();

//return;

	PPC *sppc = new PPC(*mppc);
	FrameBuffer *sfb = new FrameBuffer(60 + mcamfb->w + 20, 60, sppc->w, sppc->h);
	sfb->SetBGR(0xFFFFFFFF);
	sfb->label("Stabilized");
	sfb->show();

	int sn = 30;
	ofs = new ofstream("mydbg/geometryErrorSensitivity.txt");

#if 0
	*mppc = *sppc;
	PPC *ltsppc = new PPC(*mppc);
	*ofs << "50cm Lateral Translation Look Straight Ahead Motion" << endl;
	for (int si = 0; si < sn; si++) {
//mppc->C = mppc->C + (V3(-0.5f, 0.0f, 0.0f));
		*ofs << "frame: " << si << endl;
		mcamfb->SetBGR(0xFFFFFFFF);
		mcamfb->ClearZB(0.0f);
		tptm.RenderWireframe(mppc, mcamfb);
		lowtm.RenderPoints(mppc, mcamfb);
		hightm.RenderPoints(mppc, mcamfb);
		mcamfb->redraw();
		sfb->SetBGR(0xFFFFFFFF);
		sfb->ClearZB(0.0f);
		tptm.RenderWireframe(sppc, sfb);
		lowtm.RenderPoints(sppc, sfb);
		hightm.RenderPoints(sppc, sfb);
		AnalyzeDepthError(&tptm, &lowtm, &hightm, mppc, sppc, sfb);
		sfb->redraw();
		if (si == sn - 1)
			sfb->SaveAsTiff("mydbg/latsEx.tif");
		Fl::check();
		*ltsppc = *mppc;
		mppc->C = mppc->C + (V3(-0.5f / (sn - 1), 0.0f, 0.0f));
//break;
	}

//return;
#endif

	*ofs << "50cm Lateral Translation Look at Center Motion" << endl;
	motionColor = V3(0.0f, 1.0f, 0.0f);
	*mppc = *sppc;
	PPC *ltcppc = new PPC(*mppc);
	for (int si = 0; si < sn; si++) {
//		*ofs << "frame: " << si << endl;
		mcamfb->SetBGR(0xFFFFFFFF);
		mcamfb->ClearZB(0.0f);
		tptm.RenderWireframe(mppc, mcamfb);
		lowtm.RenderPoints(mppc, mcamfb);
		hightm.RenderPoints(mppc, mcamfb);
		mcamfb->redraw();

		sfb->SetBGR(0xFFFFFFFF);
		sfb->ClearZB(0.0f);
		tptm.RenderWireframe(sppc, sfb);
		lowtm.RenderWireframe(sppc, sfb);
		hightm.RenderWireframe(sppc, sfb);
		sfb->ClearZB(0.0f);
		AnalyzeDepthError(&tptm, &lowtm, &hightm, mppc, sppc, sfb);
		sfb->redraw();
		if (si == sn - 1) {
			sfb->SaveAsTiff("mydbg/latcEx.tif");
//			mcamfb->SaveAsTiff("mydbg/latcExM.tif");
		}
		Fl::check();
		*ltcppc = *mppc;
		mppc->C = mppc->C + (V3(-0.5f / (sn - 1), 0.0f, 0.0f));
		mppc->PositionAndOrient(mppc->C, tptm.GetCenter(), V3(0.0f, 1.0f, 0.0f));
	}

//return;

	PPC *pppc = new PPC(*mppc);
	motionColor = V3(0.0f, 0.0f, 1.0f);
	*ofs << "15deg Pan Motion with 10cm Lateral Translation" << endl;
	*mppc = *sppc;
	for (int si = 0; si < sn; si++) {
//		*ofs << "frame: " << si << endl;
		mcamfb->SetBGR(0xFFFFFFFF);
		mcamfb->ClearZB(0.0f);
		tptm.RenderWireframe(mppc, mcamfb);
		lowtm.RenderPoints(mppc, mcamfb);
		hightm.RenderPoints(mppc, mcamfb);
		mcamfb->redraw();

		sfb->SetBGR(0xFFFFFFFF);
		sfb->ClearZB(0.0f);
		tptm.RenderWireframe(sppc, sfb);
		lowtm.RenderWireframe(sppc, sfb);
		hightm.RenderWireframe(sppc, sfb);
		sfb->ClearZB(0.0f);
		AnalyzeDepthError(&tptm, &lowtm, &hightm, mppc, sppc, sfb);
		sfb->redraw();
		if (si == sn - 1) {
			sfb->SaveAsTiff("mydbg/panEx.tif");
//			mcamfb->SaveAsTiff("mydbg/panExM.tif");
		}
		Fl::check();
		*pppc = *mppc;
		mppc->Pan(25.0f / (sn - 1));
		mppc->C = mppc->C + (V3(-0.1f / (sn - 1), 0.0f, 0.0f));
	}

	ppc->C = sppc->C + V3(-1.0f, 1.5f, -3.0f);
	ppc->PositionAndOrient(ppc->C, tptm.GetCenter(), V3(0.0f, 1.0f, 0.0f));
	ppc->Tilt(5.0f);
	fb->SetBGR(0xFFFFFFFF);
	fb->ClearZB(0.0f);
	tptm.RenderWireframe(ppc, fb);
	lowtm.RenderWireframe(ppc, fb);
	hightm.RenderWireframe(ppc, fb);
	float vf = 0.25f;
//	ltsppc->Visualize(ppc, fb, vf, V3(1.0f, 0.0f, 0.0f));
	ltcppc->Visualize(ppc, fb, vf, V3(0.0f, 1.0f, 0.0f));
	fb->Draw3DPoint(tptm.GetCenter(), V3(0.0f, 0.0f, 0.0f), ppc, 5);
	fb->Draw3DSegment(ltcppc->C, V3(0.0f, 1.0f, 0.0f), tptm.GetCenter(), V3(0.0f, 1.0f, 0.0f), ppc);
	pppc->Visualize(ppc, fb, vf, V3(0.0f, 0.0f, 1.0f));
	V3 ip = pppc->C.IntersectRayWithPlane(pppc->GetVD().UnitVector(), V3(0.0f, 1.0f, 0.0f), tptm.verts[0]);
	fb->Draw3DPoint(ip, V3(0.0f, 0.0f, 1.0f), ppc, 5);
	fb->Draw3DSegment(pppc->C, V3(0.0f, 0.0f, 1.0f), ip, V3(0.0f, 0.0f, 1.0f), ppc);
	sppc->Visualize(ppc, fb, vf, V3(0.0f, 0.0f, 0.0f));
	fb->Draw3DSegment(sppc->C, V3(0.0f, 0.0f, 0.0f), tptm.GetCenter(), V3(0.0f, 0.0f, 0.0f), ppc);
	fb->SaveAsTiff("mydbg/motions.tif");
	fb->redraw();


	ofs->close();

	sppc->Pan(2.0f);
	ip = sppc->C.IntersectRayWithPlane(sppc->GetVD().UnitVector(), 
		V3(0.0f, 1.0f, 0.0f), tptm.verts[0]);
	sppc->Pan(-2.0f);
	V3 pip; sppc->Project(ip, pip);
	V3 cip(sppc->w / 2.0f, sppc->h / 2.0f, pip[2]);
	cerr << (pip-cip).Length() << endl;

	sppc->C = sppc->C - V3(0.02f, 0.0f, 0.0f);
	ip = sppc->C.IntersectRayWithPlane(sppc->GetVD().UnitVector(),
		V3(0.0f, 1.0f, 0.0f), tptm.verts[0]);
	sppc->C = sppc->C + V3(0.02f, 0.0f, 0.0f);
	sppc->Project(ip, pip);
	cip = V3(sppc->w / 2.0f, sppc->h / 2.0f, pip[2]);
	cerr << (pip - cip).Length() << endl;

}


void Scene::DBG() {

	{

		ShowPano();
		return;

	}

	{

		TestDepthSensitivity();
		return;

	}

	{
		FrameBuffer *leftfb = new FrameBuffer(30, 40, 10, 10);
		leftfb->LoadTiff("images/left.tif");
		leftfb->label("Left");
		leftfb->show();

		FrameBuffer *rightfb = new FrameBuffer(30 + 30 + leftfb->w, 40, 10, 10);
		rightfb->LoadTiff("images/right.tif");
		rightfb->label("Right");
		rightfb->show();
		return;
	}

	{
		PPC ppc0 = *ppc;
		PPC ppc1 = *ppc;
		ppc1.C = ppc1.C + V3(30.0f, 60.0f, 0.0f);
		ppc1.PositionAndOrient(ppc1.C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));
ppc1 = *ppc;
		fb->hide();
		int framesN = 1000;
		for (int i = 0; i < framesN; i++) {
			ka = (float)i / (float)framesN;
			mf = (float)i / (float)framesN;
			ppc->SetInterpolated(&ppc0, &ppc1, i, framesN);
//			Render(ppc, fb);
			hwfb->redraw();
			gpufb->redraw();
			Fl::check();
		}
		*ppc = ppc0;
		return;

	}




	{
		for (int i = 0; i < 100; i++) {
			ppc->C = ppc->C + ppc->a.UnitVector()*0.1f;
			Render(ppc, fb);
			hwfb->redraw();
			Fl::check();
		}
		return;

	}

	{

		fb->SetBGR(0xFFFFFFFF);
		fb->ClearZB(0.0f);
		tms[0].RayTrace(ppc, fb);
		fb->redraw();
		return;
	}

	{

		L = ppc->C;
		tms[0].Light(V3(1.0f, 0.0f, 0.0f), L);
		for (int i = 0; i < 360; i++) {
			kse = 1000.0f * ((float) i) / 359.0f;
			Render();
			Fl::check();
		}
		cerr << "INFO: done" << endl;
		return;

	}

	{
		V3 tc = tms[0].GetCenter();
		for (int i = 0; i < 360; i++) {
			L = tc + V3(40.0f, 0.0f, 0.0f);
			L = L.RotateThisPointAboutArbitraryAxis(tc, V3(0.0f, 1.0f, 0.0f), (float)(i*2));
			tms[0].Light(V3(1.0f, 0.0f, 0.0f), L);
			Render();
			Fl::check();
		}
		return;

	}


	{

		PPC ppc1(*ppc);

		ppc->C = V3(20.0f, 100.0f, 200.0f);
		ppc->PositionAndOrient(ppc->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));

		for (int i = 0; i < 1000; i++) {
			ppc1.Pan(.5f);
			Render();
			ppc1.Visualize(ppc, fb, 20.0f, V3(0.0f, 0.0f, 0.0f));
			Fl::check();
		}
		return;
	}


	{

		for (int i = 0; i < 30; i++) {
			ppc->Pan(.5f);
			Render();
			Fl::check();
		}
		return;


	}


	{

		V3 tmc = tms[0].GetCenter();
		V3 C1 = ppc->C + V3(120.0f, 80.0f, 0.0f);
		ppc->PositionAndOrient(C1, tmc, V3(0.0f, 1.0f, 0.0));
		Render();
		return;

	}


	{
		TM tm;
		tm.LoadBin("geometry/teapot1K.bin");
		V3 tmC = ppc->C + ppc->GetVD()*100.0f;
		float tmSize = 50.0f;
		tm.PositionAndSize(tmC, tmSize);
		fb->SetBGR(0xFFFFFFFF);
		fb->ClearZB(0.0f);
		tm.RenderWireframe(ppc, fb);
		TM tm2;
//		tm2.SetRectangle(tmC, 50.0f, 50.0f);
//		tm2.RenderWireframe(ppc, fb);
		fb->redraw();

		for (int i = 0; i < 100; i++) {
			fb->SetBGR(0xFFFFFFFF);
			fb->ClearZB(0.0f);
			ppc->ChangeFocalLength(1.05f);
			tm.RenderWireframe(ppc, fb);
			fb->redraw();
			Fl::check();
		}

		return;
	}

	{

		V3 p(20.0f, -20.0f, -80.0f);
		V3 Oa(0.0f, -20.0f, -100.0f), a(0.0f, 1.0f, 0.0f);
		fb->Draw3DSegment(Oa, V3(0.0f, 0.0f, 0.0f), Oa + a * 30.0f, V3(0.0f, 1.0f, 0.0f), ppc);
		for (int i = 0; i < 270; i++) {
			p = p.RotateThisPointAboutArbitraryAxis(Oa, a, 1.0f);
			fb->Draw3DPoint(p, V3(1.0f, 0.0f, 0.0f), ppc, 7);
			fb->redraw();
			Fl::check();
		}
		return;

	}



	{
		TM tm;
		tm.SetRectangle(V3(0.0f, 0.0f, -100.0f), 45.0f, 30.0f);
//		tm.RenderPoints(ppc, fb);
		int stepsN = 360;
		for (int stepsi = 0; stepsi < stepsN; stepsi++) {
			fb->SetBGR(0xFFFFFFFF);
			tm.RotateAboutArbitraryAxis(V3(0.0f, 0.0f, -100.0f), V3(0.0f, 1.0f, 0.0f),
				1.0f);
			tm.RenderWireframe(ppc, fb);
			fb->redraw();
			Fl::check();
		}
		return;

	}

	{

		V3 p0(-20.0f, 0.0f, -100.0f), pp;
		V3 p1(-20.0f, 0.0f, -1000.0f);
		int stepsN = 5;
		for (int stepi = 0; stepi < stepsN; stepi++) {
//			fb->SetBGR(0xFFFFFFFF);
			V3 p = p0 + (p1 - p0) * ((float)stepi / (float)(stepsN - 1));
			fb->Draw3DPoint(p, V3(1.0f, 0.0f, 0.0f), ppc, 7);
			fb->redraw();
			Fl::check();
		}
		return;

	}


	{

		V3 p0(20.1f, 40.0f, 0.0f);
		V3 p1(50.9f, 240.0f, 0.0f);
		V3 c0(1.0f, 0.0f, 0.0f);
		V3 c1(0.0f, 1.0f, 0.0f);
		fb->DrawSegment(p0, c0, p1, c1);
		fb->redraw();
		return;
	}


	{
		V3 v(1.0f, 0.0f, 0.0f);
		for (int i = 0; i < 1000; i++) {
			float scf = (float)i / 999.0f;
			v[1] = scf;
			unsigned int color = v.GetColor();
			V3 v1; v1.SetFromColor(color);
			cerr << v1 - v << "          \r";
			fb->SetBGR(color);
			fb->redraw();
			Fl::check();
		}
		cerr << endl;
		return;
	}

	{
		int u0 = 100;
		int v0 = 50;
		int u1 = 300;
		int v1 = 400;
		int r = 30;
		int stepsN = 100;
		for (int stepi = 0; stepi < stepsN; stepi++) {
			fb->SetBGR(0xFFFFFFFF);
//			fb->DrawRectangle(u0 + stepi, v0, u1 + stepi, v1, 0xFF0000FF);
			fb->DrawCircle(u0 + stepi, v0, r, 0xFF0000FF);
			fb->redraw();
			Fl::check();
		}
		return;
	}

	{
		M33 m;
		m[0] = V3(1.0f, 0.0f, 0.0f);
		m[1] = V3(0.0f, 1.0f, 0.0f);
		m[2] = V3(0.0f, 0.0f, 1.0f);
		M33 m1;
		m1[0] = V3(-1.0f, 2.0f, 3.0f);
		m1[1] = V3(-2.0f, 2.0f, 3.0f);
		m1[2] = V3(-3.0f, -2.0f, 5.0f);

		cerr << m1 << endl << m * m1 << endl << m1 * m << endl;
		return;

		cerr << m << endl;
		V3 v0(1.0f, -2.0f, 0.5f);
		cerr << m * v0;
		cerr << v0;

		return;

	}

	{
		V3 v0(1.0f, -2.0f, 0.5f);
		V3 v1(-1.0f, -4.0f, 3.0f);
		cerr << v0 << v1;
		cerr << "v0v1 = " << v0 * v1 << endl;
		return;
	}

	V3 v(1.0f, 3.0f, -1.0f);
	cerr << v[1] << endl;
	v[1] = 2.0f;
	cerr << v[1] << endl;
	return;

	int u0 = 100;
	int v0 = 50;
	int u1 = 300;
	int v1 = 400;

	fb->DrawRectangle(u0, v0, u1, v1, 0xFF0000FF);
	fb->redraw();

	fb->SaveAsTiff("mydbg/rr.tif");
	
	FrameBuffer *fb1 = new FrameBuffer(30 + fb->w, 30, fb->w, fb->h);
	fb1->LoadTiff("mydbg/rr.tif");
	fb1->label("Loaded image");
	fb1->show();

	return;

	for (int u = 0; u < fb->w; u++) {
		fb->Set(u, fb->h / 2, 0xFF000000);
	}
	fb->redraw();
	cerr << "INFO: pressed DBG" << endl;

}

void Scene::ShowPano() {

	leftfb->hide();
	rightfb->hide();

	V3 rot0(-25.0f, 0.0f, 0.0f);
	V3 rotAmps(20.0f, 10.0f, 5.0f);
	int rotStepsN[3] = { 39, 17, 15 };
	float f0 = leftppc->GetFocalLength();
	float fAmp = f0 * 0.5f;
	int fStepsN = 7;
	int stepsN = rotStepsN[0] * rotStepsN[1] * rotStepsN[2];
	V3 rotGuess;
	V3 bestGuess;
	float bestError = FLT_MAX;
	int si = 0;
	float fGuess;
	PPC ppc0(*leftppc);
	for (int fi = 0; fi < fStepsN; fi++) {
		fGuess = f0 + -fAmp / 2.0f + fAmp * (float)fi / (float)(fStepsN - 1);
		for (int i = 0; i < rotStepsN[0]; i++) {
			rotGuess[0] = rot0[0] - rotAmps[0] / 2.0f + rotAmps[0] * (float)i / (float)(rotStepsN[0] - 1);
			for (int j = 0; j < rotStepsN[1]; j++) {
				rotGuess[1] = rot0[1] - rotAmps[1] / 2.0f + rotAmps[1] * (float)j / (float)(rotStepsN[1] - 1);
				for (int k = 0; k < rotStepsN[2]; k++) {
					rotGuess[2] = rot0[2] - rotAmps[2] / 2.0f + rotAmps[2] * (float)k / (float)(rotStepsN[2] - 1);

					*rightppc = *leftppc;
					rightppc->Pan(rotGuess[0]);
					rightppc->Tilt(rotGuess[1]);
					rightppc->Roll(rotGuess[2]);

					float colorDifference = leftfb->ColorDifference(leftppc, rightfb, rightppc);
					if (bestError > colorDifference) {
						bestError = colorDifference;
						bestGuess = rotGuess;
						fb->SetBGR(0xFF000000);
						fb->Resample(ppc, leftfb, leftppc);
						fb->Resample(ppc, rightfb, rightppc);
						fb->redraw();
						Fl::check();
					}
					int pi = (int)(100.0f * (float)si / (float)stepsN);
					cerr << "best error: " << bestError << "; progress: " << pi
						<< "%           \r";
					si++;

				}
			}
		}
	}

	return;

	float panRange = 50.0f;
	float pan0 = -25.0f;
	int sn = 21;
	leftfb = rightfb;
	for (int si = 0; si < sn; si++) {
		*rightppc = *leftppc;
		float currPan = pan0 - panRange / 2.0f + panRange * (float)si / (float)sn;
		rightppc->Pan(currPan);
		fb->SetBGR(0xFF000000);
		fb->Resample(ppc, leftfb, leftppc);
		fb->Resample(ppc, rightfb, rightppc);
		float colorDifference = leftfb->ColorDifference(leftppc, rightfb, rightppc);
		cerr << "color difference: " << colorDifference << "              \r";
		fb->redraw();
		Fl::check();
	}

}